"""Classes and methods for chemical fingerprint storage and comparison.

Author: Seth Axen
E-mail: seth.axen@gmail.com"""
from __future__ import division, print_function
from collections import defaultdict
try:
    import cPickle as pkl
except ImportError:  # Python 3
    import pickle as pkl

import numpy as np
from scipy.sparse import issparse, csr_matrix
from rdkit.DataStructs.cDataStructs import ExplicitBitVect, SparseBitVect
from python_utilities.io_tools import smart_open

# ----------------------------------------------------------------------------#
# Fingerprint Classes
# ----------------------------------------------------------------------------#

BITS_DEF = 2**32
FOLD_BITS_DEF = 1024
FP_DTYPE = np.bool_
COUNT_FP_DTYPE = np.uint16
FLOAT_FP_DTYPE = np.float64


class Fingerprint(object):

    """A class to store, represent, and fold molecular fingerprints.

    Usage
    -----
    >>> import e3fp.fingerprint.fprint as fp
    >>> import numpy as np
    >>> bits = 2**32
    >>> indices1 = np.random.randint(0, bits, 100)
    >>> f1 = fp.Fingerprint(indices1, bits=bits, level=0)
    >>> f1_folded = f1.fold(bits=1024)
    >>> print("Bitstring: " + f1_folded.to_bitstring())
    Bitstring: 000000000000000000000000000000100000000100010000000001000...
    >>> indices2 = np.random.randint(0, bits, 100)
    >>> f2 = fp.Fingerprint.from_indices(indices2, bits=bits, level=0)
    >>> f2_folded = f2.fold(bits=1024)
    >>> print("Tanimoto coefficient: %.4f" % (fp.tanimoto(f1_folded,
    >>>                                                   f2_folded)))
    Tanimoto coefficient: 0.0452

    Parameters
    ----------
    indices : ndarray of int, optional (default None)
        log2(`bits`)-bit indices in a sparse bitvector of `bits` which
        correspond to 1.
    bits : int, optional (default 2**32)
        Number of bits in bitvector.
    level : int, optional (default -1)
        Level of fingerprint. 0th level just uses initial atom identifiers,
        1st level is after 1st iteration, `n`th level is after `n` iterations.
    """

    vector_dtype = FP_DTYPE

    def __init__(self, indices, bits=BITS_DEF, level=-1, props={}, **kwargs):
        """Initialize Fingerprint object."""
        self.reset()

        indices = np.asarray(indices, dtype=np.long)

        if np.any(indices >= bits):
            raise BitsValueError(
                "number of bits is lower than provided indices")

        self.indices = np.unique(indices)
        self.bits = bits
        self.level = level
        self.update_props(props)

    def clear(self):
        """Clear temporary (and possibly large) values."""
        pass

    def reset(self):
        """Reset all values."""
        self.indices = np.asarray([], dtype=np.long)
        self.bits = 0
        self.level = -1
        self.folded_fingerprint = {}
        self.index_to_folded_index_dict = None
        self.unfolded_fingerprint = None
        self.index_to_unfolded_index_dict = None
        self.props = {}
        self.clear()

    @classmethod
    def from_indices(cls, indices, bits=BITS_DEF, level=-1, **kwargs):
        """Initialize from an array of indices.

        Parameters
        ----------
        indices : ndarray of int, optional (default None)
            Indices in a sparse bitvector of length `bits` which correspond
            to 1.
        bits : int, optional (default 2**32)
            Number of bits in array. Indices will be log2(`bits`)-bit
            integers.
        level ; int, optional (default -1)
            Level of fingerprint. 0th level just uses initial atom
            identifiers, 1st level is after 1st iteration, `n`th level is
            after `n` iterations.

        Returns
        -------
        Fingerprint : fingerprint
        """
        return cls(indices, bits=bits, level=level, **kwargs)

    @classmethod
    def from_vector(cls, vector, level=-1, **kwargs):
        """Initialize from vector.

        Parameters
        ----------
        vector : ndarray or sparse matrix
            Array of bits/counts/floats
        level ; int, optional (default -1)
            Level of fingerprint. 0th level just uses initial atom
            identifiers, 1st level is after 1st iteration, `n`th level is
            after `n` iterations.

        Returns
        -------
        Fingerprint : fingerprint
        """
        if kwargs.get("bits", None) is None:
            try:
                kwargs["bits"] = vector.shape[1]
            except IndexError:
                kwargs["bits"] = vector.shape[0]
        if issparse(vector):
            indices = vector.indices.astype(np.long)
            counts = vector.data
        else:
            indices = np.asarray(np.where(vector), dtype=np.long).flatten()
            counts = vector[indices]
        counts = dict(zip(indices, counts))
        return cls.from_indices(indices, counts=counts, level=level, **kwargs)

    @classmethod
    def from_bitstring(cls, bitstring, level=-1, **kwargs):
        """Initialize from bitstring (e.g. '10010011').

        Parameters
        ----------
        bitstring : str
            String of 1s and 0s.
        level ; int, optional (default -1)
            Level of fingerprint. 0th level just uses initial atom
            identifiers, 1st level is after 1st iteration, `n`th level is
            after `n` iterations.

        Returns
        -------
        Fingerprint : fingerprint
        """
        indices = [i for i, char in enumerate(bitstring) if char != '0']
        if kwargs.get("bits", None) is None:
            kwargs["bits"] = len(bitstring)
        return cls.from_indices(indices, level=level, **kwargs)

    @classmethod
    def from_fingerprint(cls, fp, **kwargs):
        """Initialize by copying existing fingerprint.

        Parameters
        ----------
        fp : Fingerprint
            Existing fingerprint.

        Returns
        -------
        Fingerprint : fingerprint
        """
        if not isinstance(fp, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (fp.__class__.__name__))

        new_fp = cls.from_indices(fp.indices, bits=fp.bits,
                                  level=fp.level)
        new_fp.update_props(fp.props)
        new_fp.folded_fingerprint = dict([(k, v.__class__.from_fingerprint(v))
                                          for k, v
                                          in fp.folded_fingerprint.items()])
        return new_fp

    @classmethod
    def from_rdkit(cls, rdkit_fprint, **kwargs):
        """Initialize from RDKit fingerprint.

        If provided fingerprint is of length 2^32 - 1, assumes real
        fingerprint is of length 2^32.

        Parameters
        -------
        rdkit_fprint : RDKit ExplicitBitVect or SparseBitVect
            Existing RDKit fingerprint.

        Returns
        -------
        Fingerprint : fingerprint
        """
        if not (isinstance(rdkit_fprint, ExplicitBitVect) or
                isinstance(rdkit_fprint, SparseBitVect)):
            raise TypeError(
                "RDKit fingerprint must be a SparseBitVect or ExplicitBitVect")
        bits = rdkit_fprint.GetNumBits()
        if bits == 2**32 - 1:
            bits = 2**32
        indices = np.asarray(rdkit_fprint.GetOnBits(), dtype=np.long)
        return cls.from_indices(indices, bits=bits, **kwargs)

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, indices):
        self._indices = np.asarray(indices, dtype=np.long)

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, level):
        self._level = level

    @property
    def bits(self):
        return self._bits

    @bits.setter
    def bits(self, bits):
        self._bits = bits

    @property
    def props(self):
        return self._props

    @props.setter
    def props(self, props):
        self._props = props

    def get_prop(self, key):
        """Get property. If not set, raise KeyError."""
        try:
            return self.props[key]
        except:
            raise KeyError

    def set_prop(self, key, val):
        """Set property."""
        self.props[key] = val

    def update_props(self, props_dict):
        """Set multiple properties at once."""
        self.props.update(props_dict)

    @property
    def name(self):
        try:
            return self.props["Name"]
        except KeyError:
            return None

    @name.setter
    def name(self, name):
        self.props["Name"] = str(name)

    @property
    def mol(self):
        try:
            return self.props["Mol"]
        except KeyError:
            return None

    @mol.setter
    def mol(self, mol):
        self.props["Mol"] = mol

    @property
    def index_id_map(self):
        try:
            return self.props["index_id_map"]
        except:
            return None

    @index_id_map.setter
    def index_id_map(self, index_id_map):
        self.props["index_id_map"] = index_id_map

    def to_vector(self, sparse=True, dtype=None):
        """Get vector of bits/counts/floats.

        Returns
        -------
        ndarray or csr_matrix : Vector of bits/counts.floats
        """
        if dtype is None:
            dtype = self.vector_dtype

        counts = self.counts
        if sparse:
            try:
                return csr_matrix(([counts[i] for i in self.indices],
                                  ([0] * self.bit_count, self.indices)),
                                  shape=(1, self.bits), dtype=dtype)
            except ValueError:
                raise BitsValueError(
                    "Number of bits is lower than size of indices")
        else:
            bitvector = np.zeros(self.bits, dtype=dtype)
            try:
                bitvector[self.indices] = [counts[i] for i in self.indices]
                return bitvector
            except IndexError:
                raise BitsValueError(
                    "Number of bits is lower than size of indices")

    def to_bitvector(self, sparse=True):
        """Get full bitvector.

        Returns
        -------
        ndarray or csr_matrix of bool : Bitvector
        """
        return self.to_vector(sparse=sparse, dtype=FP_DTYPE)

    def to_bitstring(self):
        """Get bitstring as string of 1s and 0s.

        Returns
        -------
        str : bitstring
        """
        bitvector = self.to_bitvector(sparse=False)
        return "".join(map(str, np.asarray(bitvector, dtype=np.int)))

    def to_rdkit(self):
        """Convert to RDKit fingerprint.

        Indices may be modified to be lower than 2^32 - 1. For 32-bit
        fingerprints, this should almost never happen.

        Returns
        -------
        rdkit_fprint : RDKit ExplicitBitVect or SparseBitVect
            Convert to bitvector used for RDKit fingerprints. If `self.bits`
            is less than 10^5, `ExplicitBitVect` is used. Otherwise,
            `SparseBitVect` is used.
        """
        rdkit_fp_type = SparseBitVect
        if self.bits < 1e5:
            rdkit_fp_type = ExplicitBitVect

        # RDKit Bitvect types can't exceed 2**32 - 1 in length
        bits = self.bits % 2**32
        indices = self.indices % (2**32 - 1)

        rdkit_fprint = rdkit_fp_type(bits)
        rdkit_fprint.SetBitsFromList(indices.tolist())
        return rdkit_fprint

    @property
    def bit_count(self):
        return self.indices.shape[0]

    def get_count(self, index):
        """Return count index in fingerprint.

        Defaults to 1 if index in `self.indices`

        Returns
        -------
        int : Count of bit in fingerprint
        """
        if index in self.indices:
            return 1
        else:
            return 0

    @property
    def counts(self):
        return dict([(k, 1) for k in self.indices])

    def mean(self):
        """Return mean, i.e. proportion of "on" bits in fingerprint.

        Returns
        -------
        float : Mean
        """
        return float(self.bit_count) / self.bits

    def std(self):
        """Return standard deviation of fingerprint.

        Returns
        -------
        float : Standard deviation
        """
        mean = self.mean()
        return (mean * (1 - mean))**0.5

    # Folding/unfolding to a new fingerprint
    def fold(self, bits=FOLD_BITS_DEF, method=0, linked=True):
        """Return fingerprint for bitvector folded to size `bits`.

        Parameters
        ----------
        bits : int, optional (default 1024)
            Length of new bitvector, ideally multiple of 2.
        method : int, optional (default 0)
            Method to use for folding.
            0: partitioning (array is divided into equal sized arrays of
               length `bits` which are bitwise combined with OR)
            1: compression (adjacent bits pairs are combined with OR until
               length is `bits`)
        linked : bool, optional (default True)
            Link folded and unfolded fingerprints for easy referencing. Set
            to False if intending to save and want to reduce file size.

        Returns
        -------
        Fingerprint : Fingerprint of folded bitvector
        """
        if bits > self.bits:
            raise BitsValueError("folded bits more greater than existing bits")
        if not np.log2(self.bits / bits).is_integer():
            raise BitsValueError(
                "existing bits divided by power of 2 does not give folded bits"
            )
        if method not in (0, 1):
            raise OptionError("method must be 0 or 1")

        if (bits, method) not in self.folded_fingerprint:
            if method == 0:
                folded_indices = self.indices % bits
            elif method == 1:
                folded_indices = self.indices / (self.bits / bits)

            self.index_to_folded_index_dict = dict(zip(self.indices,
                                                       folded_indices))
            folded_index_to_index_dict = {}
            for index, folded_index in self.index_to_folded_index_dict.items():
                folded_index_to_index_dict.setdefault(folded_index,
                                                      set([])).add(index)

            fp = self.__class__.from_indices(folded_indices, bits=bits,
                                             level=self.level)
            fp.update_props(self.props)

            fp.index_to_unfolded_index_dict = folded_index_to_index_dict
            if self.index_id_map is not None:
                fp.index_id_map = {}
                for index, id_set in self.index_id_map.items():
                    fp.index_id_map.setdefault(
                        self.index_to_folded_index_dict[index],
                        set()).update(id_set)

            if linked:
                fp.unfolded_fingerprint = self
                self.folded_fingerprint[(bits, method)] = fp

        assert isinstance(self.folded_fingerprint[(bits, method)],
                          self.__class__)
        return self.folded_fingerprint[(bits, method)]

    def get_folding_index_map(self):
        """Get map of sparse indices to folded indices.

        Returns
        -------
        dict : Map of sparse index (keys) to corresponding folded index.
        """
        return self.index_to_folded_index_dict

    def unfold(self):
        """Return unfolded parent fingerprint for bitvector.

        Returns
        -------
        Fingerprint : Fingerprint of unfolded bitvector. If None, return
                      None.
        """
        return self.unfolded_fingerprint

    def get_unfolding_index_map(self):
        """Get map of sparse indices to unfolded indices.

        Returns
        -------
        dict : Map of sparse index (keys) to set of corresponding unfolded
               indices.
        """
        return self.index_to_unfolded_index_dict

    # summary magic methods
    def __repr__(self):
        return "%s(indices=%s, level=%r, bits=%r, name=%s)" % (
            self.__class__.__name__,
            repr(self.indices
                 ).replace('\n', '').replace(' ', '').replace(',', ', '),
            self.level, self.bits, self.name)

    def __str__(self):
        return self.__repr__()

    # logical/comparative magic methods
    def __eq__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        return (self.level == other.level
                and self.bits == other.bits
                and self.__class__ == other.__class__
                and np.all(np.in1d(self.indices, other.indices,
                           assume_unique=True)))

    def __ne__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        return not self.__eq__(other)

    def __add__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        if self.bits != other.bits:
            raise BitsValueError("cannot add fingerprints of different sizes")

        return Fingerprint(np.union1d(self.indices, other.indices),
                           bits=self.bits)

    def __sub__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        if self.bits != other.bits:
            raise BitsValueError(
                "cannot subtract fingerprints of different sizes")

        return Fingerprint(np.setdiff1d(self.indices,
                                        other.indices,
                                        assume_unique=True),
                           bits=self.bits)

    def __and__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        if self.bits != other.bits:
            raise BitsValueError(
                "cannot compare fingerprints of different sizes")

        return Fingerprint(np.intersect1d(self.indices,
                                          other.indices,
                                          assume_unique=True),
                           bits=self.bits)

    def __or__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        if self.bits != other.bits:
            raise BitsValueError(
                "cannot compare fingerprints of different sizes")

        return Fingerprint(np.union1d(self.indices, other.indices),
                           bits=self.bits)

    def __xor__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (other.__class__.__name__))

        if self.bits != other.bits:
            raise BitsValueError(
                "cannot compare fingerprints of different sizes")

        return Fingerprint(np.setxor1d(self.indices, other.indices,
                                       assume_unique=True),
                           bits=self.bits)

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __rand__(self, other):
        return self.__and__(other)

    def __ror__(self, other):
        return self.__ror__(other)

    def __rxor__(self, other):
        return self.__rxor__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __isub__(self, other):
        return self.__sub__(other)

    def __iand__(self, other):
        return self.__and__(other)

    def __ior__(self, other):
        return self.__ror__(other)

    def __ixor__(self, other):
        return self.__rxor__(other)

    # iterable magic methods
    def __len__(self):
        return self.bits

    def __getitem__(self, key):
        if type(key) is not int:
            raise TypeError
        elif self.indices is None:
            raise KeyError
        elif key > self.bits:
            raise KeyError
        elif key < -self.bits:
            raise KeyError
        else:
            return (key in self.indices)

    # pickle magic methods, reduces size of fingerprint file
    def __getstate__(self):
        return dict([(k, v) for k, v in self.__dict__.items()])

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.clear()


class CountFingerprint(Fingerprint):

    """A fingerprint that stores number of occurrences of each index.

    Usage
    -----
    >>> import e3fp.fingerprint.fprint
    >>> import numpy as np
    >>> bits = 2**32
    >>> indices1 = np.unique(np.random.randint(0, bits, 100))
    >>> counts1 = dict(zip(indices1, np.random.randint(1, 100,
    >>>                                                indices1.shape[0])))
    >>> f1 = fp.CountFingerprint(indices1, bits=bits, counts=counts1, level=0)
    >>> f1_folded = f1.fold(bits=1024)
    >>> indices2 = np.random.randint(0, bits, 10)
    >>> f2 = fp.CountFingerprint.from_indices(indices2, bits=bits, level=0)
    >>> f2_folded = f2.fold(bits=1024)
    >>> print("Soergel similarity: %.4f" % (fp.soergel_comp(f1_folded,
    >>>                                                     f2_folded)))
    Soergel similarity: 0.0002
    >>> f3_folded = f1_folded + f2_folded
    >>> print("Soergel similarity: %.4f" % (fp.soergel_comp(f1_folded,
    >>>                                                     f3_folded)))
    Soergel similarity: 0.9980
    """

    vector_dtype = COUNT_FP_DTYPE

    def __init__(self, indices=None, counts=None, bits=BITS_DEF, level=-1,
                 props={}, **kwargs):
        """Initialize CountFingerprint object."""
        if indices is None and counts is None:
            raise TypeError("indices or counts must be specified")

        self.reset()

        if indices is not None:
            indices = np.asarray(indices, dtype=np.long)

            if np.any(indices >= bits):
                raise BitsValueError(
                    "number of bits is lower than provided indices")

            if counts is None:
                indices, counts = np.unique(indices, return_counts=True)
                counts = dict(zip(indices, counts))
            else:
                indices = np.unique(indices)
                if not np.all([x in indices for x in counts]):
                    raise CountsError(
                        "At least one index in `counts` is not in `indices`.")
                if len(set(indices).symmetric_difference(counts)) > 0:
                    raise CountsError(
                        "At least one index in `indices` is not in `counts`.")

        else:
            indices = np.asarray(sorted(counts.keys()), dtype=np.long)

            if np.any(indices >= bits):
                raise BitsValueError(
                    "number of bits is lower than provided indices")

        self.indices = indices
        self.counts = counts
        self.bits = bits
        self.level = level
        self.update_props(props)

    @classmethod
    def from_indices(cls, indices, counts=None, bits=BITS_DEF, level=-1,
                     **kwargs):
        """Initialize from an array of indices.

        Parameters
        ----------
        indices : ndarray of int, optional (default None)
            Indices in a sparse bitvector of length `bits` which correspond to
            1.
        counts : dict, optional (default None)
            Dictionary mapping sparse indices to counts.
        bits : int, optional (default 2**32)
            Number of bits in array. Indices will be log2(`bits`)-bit
            integers.
        level : int, optional (default -1)
            Level of fingerprint. 0th level just uses initial atom
            identifiers, 1st level is after 1st iteration, `n`th level is
            after `n` iterations.

        Returns
        -------
        CountFingerprint : fingerprint
        """
        return cls(indices, counts=counts, bits=bits, level=level, **kwargs)

    @classmethod
    def from_counts(cls, counts, bits=BITS_DEF, level=-1, **kwargs):
        """Initialize from an array of indices.

        Parameters
        ----------
        counts : dict
            Dictionary mapping sparse indices to counts.
        bits : int, optional (default 2**32)
            Number of bits in array. Indices will be log2(`bits`)-bit
            integers.
        level : int, optional (default -1)
            Level of fingerprint. 0th level just uses initial atom
            identifiers, 1st level is after 1st iteration, `n`th level is
            after `n` iterations.

        Returns
        -------
        CountFingerprint : fingerprint
        """
        return cls(counts=counts, bits=bits, level=level, **kwargs)

    @classmethod
    def from_fingerprint(cls, fp, **kwargs):
        """Initialize by copying existing fingerprint.

        Parameters
        ----------
        fp : Fingerprint
            Existing fingerprint.

        Returns
        -------
        Fingerprint : fingerprint
        """
        if not isinstance(fp, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not Fingerprint" % (fp.__class__.__name__))

        counts = dict([(i, c) for i, c in fp.counts.items() if c > 0])
        new_fp = cls.from_counts(counts, bits=fp.bits,
                                 level=fp.level)
        new_fp.update_props(fp.props)
        new_fp.folded_fingerprint = dict([(k, v.__class__.from_fingerprint(v))
                                          for k, v
                                          in fp.folded_fingerprint.items()])
        return new_fp

    def reset(self, *args, **kwargs):
        """Reset all values."""
        super(CountFingerprint, self).reset(*args, **kwargs)
        self.counts = {}

    def get_count(self, index):
        """Return count index in fingerprint.

        Returns
        -------
        int : Count of index in fingerprint
        """
        return self.counts.get(index, 0)

    @property
    def counts(self):
        return self._counts

    @counts.setter
    def counts(self, counts):
        self._counts = dict([(k, int(v)) for k, v in counts.items()])

    def fold(self, *args, **kwargs):
        """Fold fingerprint while considering counts.

        Optionally, provide a function to reduce colliding counts.

        Parameters
        ----------
        bits : int, optional (default 1024)
            Length of new bitvector, ideally multiple of 2.
        method : int, optional (default 0)
            Method to use for folding.
            0: partitioning (array is divided into equal sized arrays of
               length `bits` which are bitwise combined with OR)
            1: compression (adjacent bits pairs are combined with OR until
               length is `bits`)
        linked : bool, optional (default True)
            Link folded and unfolded fingerprints for easy referencing. Set to
            False if intending to save and want to reduce file size.
        counts_method : function, optional (default sum)
            Function for combining counts. Default is summation.

        Returns
        -------
        CountFingerprint : Fingerprint of folded bitvector
        """
        counts_method = kwargs.get("counts_method", sum)

        fp = super(CountFingerprint, self).fold(*args, **kwargs)
        counts = dict([(fold_ind, counts_method([self.get_count(x)
                                                 for x in ind_set]))
                       for fold_ind, ind_set
                       in fp.index_to_unfolded_index_dict.items()])
        fp.counts = counts
        return fp

    # summary magic methods
    def __repr__(self):
        return "%s(counts=%r, level=%r, bits=%r, name=%s)" % (
            self.__class__.__name__, self.counts, self.level,
            self.bits, self.name)

    # logical/comparative magic methods
    def __eq__(self, other):
        if not isinstance(other, CountFingerprint):
            raise InvalidFingerprintError(
                "variable is %s not CountFingerprint" % (
                    other.__class__.__name__))

        return (self.level == other.level
                and self.bits == other.bits
                and self.counts == other.counts
                and self.__class__ == other.__class__)

    def __ne__(self, other):
        if not isinstance(other, Fingerprint):
            raise InvalidFingerprintError(
                "variable is %s not CountFingerprint" % (
                    other.__class__.__name__))

        return not self.__eq__(other)

    def __add__(self, other):
        if not isinstance(other, CountFingerprint):
            raise InvalidFingerprintError("variable is not CountFingerprint.")

        if self.bits != other.bits:
            raise BitsValueError("cannot add fingerprints of different sizes")

        if self.level == other.level:
            level = self.level
        else:
            level = -1

        new_counts = self.counts.copy()
        for k, v in list(other.counts.items()):
            new_counts[k] = new_counts.get(k, 0) + v

        new_indices = np.asarray(list(new_counts.keys()), dtype=np.long)

        if other.__class__ is FloatFingerprint:
            new_class = FloatFingerprint
        else:
            new_class = self.__class__

        return new_class(new_indices, counts=new_counts, bits=self.bits,
                         level=level)

    def __sub__(self, other):
        if not isinstance(other, CountFingerprint):
            raise InvalidFingerprintError("variable is not CountFingerprint.")

        if self.bits != other.bits:
            raise BitsValueError(
                "cannot subtract fingerprints of different sizes")

        if self.level == other.level:
            level = self.level
        else:
            level = -1

        new_counts = self.counts.copy()
        for k, v in other.counts.items():
            new_counts[k] = new_counts.get(k, 0) - v

        new_indices = np.asarray(new_counts.keys(), dtype=np.long)

        if other.__class__ is FloatFingerprint:
            new_class = FloatFingerprint
        else:
            new_class = self.__class__

        return new_class(new_indices, counts=new_counts, bits=self.bits,
                         level=level)

    def __floordiv__(self, x):
        cf = CountFingerprint.from_fingerprint(self)
        cf.counts = dict([(k, int(v / x)) for k, v in self.counts.items()
                          if v >= x])
        return cf

    def __div__(self, x):
        x = float(x)
        cf = FloatFingerprint.from_fingerprint(self)
        cf.counts = dict([(k, v / x) for k, v in self.counts.items()])
        return cf

    def __truediv__(self, x):
        return self.__div__(x)

    def __mul__(self, x):
        cf = self.__class__.from_fingerprint(self)
        cf.counts = dict([(k, v * float(x)) for k, v in self.counts.items()])
        return cf

    def __rfloordiv__(self, x):
        return self.__floordiv__(x)

    def __rdiv__(self, x):
        return self.__div__(x)

    def __rtruediv__(self, x):
        return self.__truediv__(x)

    def __rmul__(self, x):
        return self.__mul__(x)

    def __ifloordiv__(self, x):
        return self.__floordiv__(x)

    def __idiv__(self, x):
        return self.__div__(x)

    def __itruediv__(self, x):
        return self.__truediv__(x)

    def __imul__(self, x):
        return self.__mul__(x)

    # iterable magic mathods
    def __len__(self):
        return self.bits

    def __getitem__(self, key):
        if type(key) is not int:
            raise TypeError
        elif self.indices is None:
            raise KeyError
        elif key > self.bits:
            raise KeyError
        elif key < -self.bits:
            raise KeyError
        else:
            return (key in self.indices)

    # pickle magic methods, reduces size of fingerprint
    def __getstate__(self):
        return dict([(k, v)
                     for k, v in self.__dict__.items()
                     if k not in ("indices",)])

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.indices = sorted(self.counts.keys())
        self.clear()


class FloatFingerprint(CountFingerprint):

    """A Fingerprint that stores float counts.

    Nearly identical to CountFingerprint. Mainly a naming convention, but
    counts are stored as floats."""

    vector_dtype = FLOAT_FP_DTYPE

    @property
    def counts(self):
        return self._counts

    @counts.setter
    def counts(self, counts):
        self._counts = dict([(k, float(v)) for k, v in counts.items()])


# ----------------------------------------------------------------------------#
# Exception Classes
# ----------------------------------------------------------------------------#


class InvalidFingerprintError(TypeError):

    """Fingerprint is incorrectly formatted."""

    pass


class MolError(TypeError):

    """Mol is of incorrect type."""

    pass


class BitsValueError(ValueError):

    """Bits value is invalid."""

    pass


class CountsError(ValueError):

    """Index in Counts is invalid."""

    pass


class OptionError(ValueError):

    """Option provided is invalid."""

    pass


# ----------------------------------------------------------------------------#
# Serialization Methods
# ----------------------------------------------------------------------------#


def load(f, update_structure=True):
    """Load ``Fingerprint`` object from file.

    Parameters
    ----------
    f : str or File
        File name or file-like object to load file from.
    update_structure : bool, optional (default True)
        Attempt to update the class structure by initializing a new, shiny
        fingerprint from each fingerprint in the file. Useful for guaranteeing
        that old, dusty fingerprints are always upgradeable.

    Returns
    -------
    Fingerprint : Pickled fingerprint.
    """
    fps = _load(f, update_structure)
    if len(fps) == 0:
        return None
    else:
        return fps[0]


def loadz(f, update_structure=True):
    """Load ``Fingerprint`` objects from file.

    Parameters
    ----------
    f : str or File
        File name or file-like object to load file from.
    update_structure : bool, optional (default True)
        Attempt to update the class structure by initializing a new, shiny
        fingerprint from each fingerprint in the file. Useful for guaranteeing
        that old, dusty fingerprints are always upgradeable. If this doesn't
        work, falls back to the original saved fingerprint.

    Returns
    -------
    list of Fingerprint : List of all fingerprints in pickle.
    """
    return _load(f, update_structure)


def _load(f, update_structure=True):
    fps = []
    with smart_open(f, "rb") as fh:
        try:
            while True:
                fp = pkl.load(fh)
                if update_structure:
                    try:
                        fps.append(fp.__class__.from_fingerprint(fp))
                    except:
                        fps.append(fp)
                else:
                    fps.append(fp)
        except EOFError:
            pass

    return fps


def save(f, fp, **kwargs):
    """Save ``Fingerprint`` object to file.

    Parameters
    ----------
    f : str or File
        filename ``str`` or file-like object to save file to
    fp : Fingerprint
        Fingerprint to save to file
    protocol : int, optional (default None)
        Pickle protocol to use. Valid options are 0, 1, or 2. If None, highest
        available protocol is used. This will not affect fingerprint loading.

    Returns
    ----------
    bool : Success or fail
    """
    return _save(f, fp, **kwargs)


def savez(f, *fps, **kwargs):
    """Save multiple ``Fingerprint`` objects to file.

    Parameters
    ----------
    f : str or File
        filename ``str`` or file-like object to save file to
    fps : list of Fingerprint
        List of Fingerprints to save to file
    protocol : int, optional (default None)
        Pickle protocol to use. Valid options are 0, 1, or 2. If None, highest
        available protocol is used. This will not affect fingerprint loading.

    Returns
    ----------
    bool : Success or fail
    """
    return _save(f, *fps, **kwargs)


def _save(f, *fps, **kwargs):
    default_dict = {'protocol': None}
    default_dict.update(kwargs)
    protocol = default_dict["protocol"]

    with smart_open(f, "wb") as fh:
        if protocol is None:
            protocol = pkl.HIGHEST_PROTOCOL

        for fp in fps:
            pkl.dump(fp, fh, protocol)

    return True


# ----------------------------------------------------------------------------#
# Fingerprint Comparison Methods
# ----------------------------------------------------------------------------#


def tanimoto(fp1, fp2):
    """Calculate Tanimoto coefficient between fingerprints.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : Tanimoto coefficient.
    """
    if not isinstance(fp1, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp1.__class__.__name__))
    if not isinstance(fp2, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp2.__class__.__name__))

    try:
        intersect = np.intersect1d(fp1.indices, fp2.indices,
                                   assume_unique=True).shape[0]
        return intersect / (fp1.bit_count + fp2.bit_count -
                            intersect)
    except ZeroDivisionError:
        return 0.


def dice(fp1, fp2):
    """Calculate Dice coefficient between fingerprints.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : Dice coefficient.
    """
    if not isinstance(fp1, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp1.__class__.__name__))
    if not isinstance(fp2, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp2.__class__.__name__))

    try:
        intersect = np.intersect1d(fp1.indices, fp2.indices,
                                   assume_unique=True).shape[0]
        return 2 * intersect / (fp1.bit_count + fp2.bit_count)
    except ZeroDivisionError:
        return 0.


def hamming(fp1, fp2):
    """Calculate Hamming distance between fingerprints.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : Hamming distance.
    """
    if not isinstance(fp1, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp1.__class__.__name__))
    if not isinstance(fp2, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp2.__class__.__name__))

    intersect = np.intersect1d(fp1.indices, fp2.indices,
                               assume_unique=True).shape[0]
    return fp1.bit_count + fp2.bit_count - 2 * intersect


def distance(fp1, fp2):
    """Calculate Euclidean distance between fingerprints.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : Euclidian distance.
    """
    return hamming(fp1, fp2)**0.5


def pearson(fp1, fp2):
    """Calculate Pearson correlation between fingerprints.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : Pearson correlation.
    """
    if not isinstance(fp1, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp1.__class__.__name__))
    if not isinstance(fp2, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp2.__class__.__name__))

    intersect = np.intersect1d(fp1.indices, fp2.indices,
                               assume_unique=True).shape[0]

    return ((intersect / fp1.bits) -
            ((fp1.mean() * fp2.mean()) / (fp1.std() * fp2.std())))


def soergel(fp1, fp2):
    """Calculate Soergel distance between fingerprints.

    For Fingerprint comparison, this is equivalent to complement of Tanimoto
    coefficient.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : Soergel distance.

    Reference
    -------

    """
    if not isinstance(fp1, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp1.__class__.__name__))
    if not isinstance(fp2, Fingerprint):
        raise InvalidFingerprintError("variable is %s not Fingerprint" % (
            fp2.__class__.__name__))

    if not (isinstance(fp1, CountFingerprint) and
            isinstance(fp2, CountFingerprint)):
        return 1. - tanimoto(fp1, fp2)

    counts_diff = diff_counts_dict(fp1, fp2)
    temp = np.asarray(
        [(abs(counts_diff[x]), max(fp1.get_count(x), fp2.get_count(x)))
         for x in counts_diff.keys()], dtype=np.float).T
    soergel = np.sum(temp[0, :]) / np.sum(temp[1, :])

    return soergel


def soergel_comp(fp1, fp2):
    """Calculate complement of Soergel coefficient between fingerprints.

    For Fingerprint comparison, this is equivalent to Tanimoto coefficient.
    For CountFingerprint comparison, this diverges.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint 1
    fp2 : Fingerprint
        Fingerprint 2

    Returns
    -------
    float : complement of Soergel coefficient.

    Reference
    -------

    """
    return 1. - soergel(fp1, fp2)


def add(fprints, weights=None):
    """Add fingerprints by count to new CountFingerprint or FloatFingerprint.

    If any of the fingerprints are FloatFingerprint, resulting fingerprint is
    likewise a FloatFingerprint. Otherwise, resulting fingerprint is
    CountFingerprint.

    Parameters
    ----------
    fprints : iterable of Fingerprint
        Fingerprints to be added by count.
    weights : iterable of float
        Weights for weighted sum. Results in `FloatFingerprint` output.

    Returns
    -------
    CountFingerprint or FloatFingerprint : Fingerprint with counts as sum of
                                           counts in `fprints`.
    """
    if len(fprints) == 0:
        return None

    if weights is None:
        new_counts = sum_counts_dict(*fprints)
        for fprint in fprints:
            if isinstance(fprint, FloatFingerprint):
                new_class = FloatFingerprint
                break
        else:
            new_class = CountFingerprint
    elif len(weights) != len(fprints):
        raise ValueError("Number of fingerprints and weights must be the same.")
    else:
        new_counts = sum_counts_dict(*fprints, weights=weights)
        new_class = FloatFingerprint

    new_indices = np.asarray(sorted(new_counts.keys()), dtype=np.long)

    return new_class(new_indices, counts=new_counts,
                     bits=fprints[0].bits,
                     level=fprints[0].level)


def mean(fprints, weights=None):
    """Average fingerprints to generate FloatFingerprint.

    Parameters
    ----------
    fprints : iterable of Fingerprint
        Fingerprints to be added by count.
    weights : iterable of float
        Weights for weighted mean. Weights are normalized to a sum of 1.

    Returns
    -------
    FloatFingerprint : Fingerprint with float counts as average of counts in
                       `fprints`.
    """
    if weights is not None:
        weights = np.asarray(weights)
        weight_sum = np.sum(weights)
        if weight_sum == 0.:
            raise ValueError("Sum of weights is 0.")
        weights = weights / weight_sum
        return add(fprints, weights=weights)
    else:
        return add(fprints) / len(fprints)


def sum_counts_dict(*fprints, **kwargs):
    """Given fingerprints, returns sum of their counts dicts.

    If an optional `weights` iterable of the same length as `fprints` is
    provided, the weighted sum is returned.

    Parameters
    ----------
    *fprints
        One or more Fingerprint objects

    Returns
    -------
    dict : Dict of non-zero count indices in any of the `fprints` with value
           as sum of counts.
    """
    counts_sum = defaultdict(int)
    if "weights" not in kwargs:
        for fprint in fprints:
            for k, v in fprint.counts.items():
                counts_sum[k] += v
    else:
        weights = kwargs["weights"]
        for (fprint, weight) in zip(fprints, weights):
            for k, v in fprint.counts.items():
                counts_sum[k] += v * weight
    return counts_sum


def diff_counts_dict(fp1, fp2, only_positive=False):
    """Given two fingerprints, returns difference of their counts dicts.

    Parameters
    ----------
    fp1 : Fingerprint
        Fingerprint object.
    fp2 : Fingerprint
        Fingerprint object to be subtracted.
    only_positive : bool, optional (default False)
        If True, only positive counts are returned

    Returns
    -------
    dict : Dict of count indices in any of the `fprints` with value as diff of
           counts.
    """
    counts_diff = fp1.counts.copy()
    for k, v in fp2.counts.items():
        counts_diff[k] = counts_diff.get(k, 0) - v
        if only_positive and counts_diff[k] < 0:
            del(counts_diff[k])
    return counts_diff



"""Utilities for handling SMILES strings and RDKit mols and conformers.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import os
import re
import copy
import logging
from collections import namedtuple

import rdkit
import rdkit.Chem
import rdkit.Chem.PropertyMol
from rdkit.Chem.PropertyMol import PropertyMol
from python_utilities.io_tools import smart_open, touch_dir


PROTO_NAME_DELIM = "-"
CONF_NAME_DELIM = "_"
MOL_ITEM_REGEX = re.compile(
    r"(?P<{0}>.+?)(?:{1}(?P<{2}>\d+))?(?:{3}(?P<{4}>\d+))?$".format(
        "mol_name", PROTO_NAME_DELIM, "proto_state_num", CONF_NAME_DELIM,
        "conf_num"))
MOL_ITEM_FIELDS = ("mol_name", "proto_state_num", "conf_num")


MolItemTuple = namedtuple("MolItemTuple", ["mol_name", "proto_state_num",
                          "conf_num"])


class MolItemName(object):
    """Class for parsing mol item names and converting to various formats."""

    def __init__(self, mol_name=None, proto_state_num=None,
                 conf_num=None, proto_delim=PROTO_NAME_DELIM,
                 conf_delim=CONF_NAME_DELIM):
        self.mol_name = mol_name
        self.proto_state_num = proto_state_num
        self.conf_num = conf_num
        self.proto_delim = proto_delim
        self.conf_delim = conf_delim

    @classmethod
    def from_str(cls, mol_item_name, mol_item_regex=MOL_ITEM_REGEX,
                 mol_item_fields=MOL_ITEM_FIELDS, **kwargs):
        fields = cls.mol_item_name_to_dict(mol_item_name,
                                           mol_item_regex=mol_item_regex,
                                           mol_item_fields=mol_item_fields)
        return cls(fields["mol_name"], fields["proto_state_num"],
                   fields["conf_num"], **kwargs)

    def to_str(self):
        return self.mol_item_name

    @classmethod
    def from_tuple(cls, fields_tuple):
        return cls(*fields_tuple)

    def to_tuple(self):
        return MolItemTuple(self.mol_name, self.proto_state_num,
                            self.conf_num)

    @property
    def mol_name(self):
        return self._mol_name

    @mol_name.setter
    def mol_name(self, mol_name):
        self._mol_name = mol_name

    def to_mol_name(self, as_proto=False):
        if as_proto:
            return self.proto_name
        else:
            return self.mol_name

    @property
    def proto_name(self):
        return self.to_proto_name(self.proto_state_num)

    def to_proto_name(self, proto_state_num=None,
                      proto_delim=PROTO_NAME_DELIM):
        if proto_state_num is not None:
            return "{}{}{:d}".format(self.mol_name, proto_delim,
                                     proto_state_num)
        else:
            return self.mol_name

    @property
    def conf_name(self):
        return self.to_conf_name(conf_num=self.conf_num)

    def to_conf_name(self, conf_num=None, conf_delim=CONF_NAME_DELIM):
        if conf_num is not None:
            return "{}{}{:d}".format(self.proto_name, conf_delim, conf_num)
        else:
            return self.proto_name

    @property
    def mol_item_name(self):
        return self.conf_name

    @staticmethod
    def mol_item_name_to_dict(mol_item_name, mol_item_regex=MOL_ITEM_REGEX,
                              mol_item_fields=MOL_ITEM_FIELDS):
        match = re.match(mol_item_regex, mol_item_name)
        groups = match.groups()
        fields = dict(zip(mol_item_fields, groups))
        proto_state_num = fields.get("proto_state_num")
        if proto_state_num is not None:
            fields["proto_state_num"] = int(proto_state_num)
        conf_num = fields.get("conf_num")
        if conf_num is not None:
            fields["conf_num"] = int(conf_num)
        return fields

    def copy(self):
        return copy.copy(self)

    def __repr__(self):
        return ("MolItemName(mol_name={}, proto_state_num={}, "
                "conf_num={})".format(self.mol_name, self.proto_state_num,
                                      self.conf_num)
                )

    def __str__(self):
        return self.conf_name

    def __eq__(self, other):
        return self.to_tuple() == other.to_tuple()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return self.to_tuple().__gt__(other.to_tuple())

    def __lt__(self, other):
        return self.to_tuple().__lt__(other.to_tuple())

    def __hash__(self):
        return hash(self.to_tuple())


def smiles_generator(*filenames):
    """Parse SMILES file(s) and yield (name, smile).

    Parameters
    ----------
    files : iterable object
        List of files containing smiles. File must contain one smile per
        line, followed by a space and then the molecule name.

    Yields
    ------
    tuple:
        ``tuple`` of the format (smile, name).
    """
    for filename in filenames:
        with smart_open(filename, "rb") as f:
            for i, line in enumerate(f):
                values = line.rstrip('\r\n').split()
                if len(values) >= 2:
                    yield tuple(values[:2])
                else:
                    logging.warning(
                        ("Line {:d} of {} has {:d} entries. Expected at least"
                         " 2.".format(i + 1, filename, len(values))),
                        exc_info=True)


def smiles_to_dict(smiles_file, unique=False, has_header=False):
    """Read SMILES file to dict."""
    smiles_gen = smiles_generator(smiles_file)
    if has_header:
        header = next(smiles_gen)
        logging.info("Skipping first (header) values: {!r}".format(header))
    if unique:
        used_smiles = set()
        smiles_dict = {}
        for smiles, name in smiles_gen:
            if name not in smiles_dict and smiles not in used_smiles:
                smiles_dict[name] = smiles
                used_smiles.add(smiles)
    else:
        smiles_dict = {name: smiles for smiles, name in smiles_gen}
    return smiles_dict


def dict_to_smiles(smiles_file, smiles_dict):
    """Write SMILES dict to file."""
    iter_to_smiles(smiles_file, sorted(smiles_dict.items()))


def iter_to_smiles(smiles_file, smiles_iter):
    """Write iterator of (mol_name, SMILES) to file."""
    with smart_open(smiles_file, "wb") as f:
        for mol_name, smiles in smiles_iter:
            f.write("{} {}\n".format(smiles, mol_name))


def mol2_generator(*filenames):
    """Parse name from mol2 filename and return generator.

    Parameters
    ----------
    files : iterable object
        List of mol2 files, where filename should be molecule name followed by
        ".mol2"

    Yields
    ------
    tuple:
        ``tuple`` of the format (file, name).
    """
    for filename in filenames:
        name = os.path.splitext(os.path.basename(filename))[0]
        yield (filename, name)


def mol_from_smiles(smiles, name, standardise=False):
    """Generate a n RDKit ``PropertyMol`` from SMILES string.

    Parameters
    ----------
    smile : str
        SMILES string
    name : str
        Name of molecule
    standardise : bool
        Clean Mol through standardisation

    Returns
    ----------
    RDKit PropertyMol : Molecule.
    """
    mol = rdkit.Chem.MolFromSmiles(smiles)
    if mol is None:
        logging.error("Mol creation failed from SMILES: {!r}".format(
            (smiles, name)))
        return None
    if standardise:
        mol = mol_to_standardised_mol(mol, name)
    mol = PropertyMol(mol)
    mol.SetProp("_Name", name)
    mol.SetProp("_SMILES", smiles)
    return mol


def mol_from_mol2(mol2_file, name=None, standardise=False):
    """Read a mol2 file into an RDKit ``PropertyMol``.

    Parameters
    ----------
    mol2_file : str
        path to a mol2 file
    name : str, optional
        Name of molecule. If not provided, uses file basename as name
    standardise : bool
        Clean mol through standardisation

    Returns
    ----------
    RDKit PropertyMol : Molecule.
    """
    if name is None:
        name = os.path.splitext(os.path.basename(mol2_file))[0]
    mol = rdkit.Chem.MolFromMol2File(mol2_file)
    if standardise:
        mol = mol_to_standardised_mol(mol, name)
    mol = PropertyMol(mol)
    mol.SetProp("_Name", name)
    return mol


def mol_from_sdf(sdf_file, conf_num=None, standardise=False):
    """Read SDF file into an RDKit ``Mol`` object.

    Parameters
    ----------
    sdf_file : str
        Path to an SDF file
    conf_num : int or None, optional
        Maximum number of conformers to read from file. Defaults to all.
    standardise : bool (default False)
        Clean mol through standardisation

    Returns
    -------
    RDKit Mol : ``Mol`` object with each molecule in SDF file as a conformer
    """
    mol = None
    with smart_open(sdf_file, "rb") as f:
        supplier = rdkit.Chem.ForwardSDMolSupplier(f)
        i = 0
        while True:
            if i == conf_num:
                break
            try:
                new_mol = next(supplier)
            except StopIteration:
                logging.debug(
                    "Read {:d} conformers from {}.".format(i, sdf_file))
                break
            if mol is None:
                mol = rdkit.Chem.Mol(new_mol)
                mol.RemoveAllConformers()
            conf = new_mol.GetConformers()[0]
            mol.AddConformer(conf, assignId=True)
            i += 1
    if standardise:
        mol = mol_to_standardised_mol(mol)
    try:
        mol.GetProp("_Name")
    except:
        name = os.path.basename(sdf_file).split(".sdf")[0]
        mol.SetProp("_Name", name)
    return mol


def mol_to_sdf(mol, out_file, conf_num=None):
    """Write RDKit ``Mol`` objects to an SDF file.

    Parameters
    ----------
    mol : RDKit Mol
        A molecule containing 1 or more conformations to write to file.
    out_file : str
        Path to save SDF file.
    conf_num : int or None, optional
        Maximum number of conformers to save to file. Defaults to all.
    """
    touch_dir(os.path.dirname(out_file))
    with smart_open(out_file, "wb") as fobj:
        writer = rdkit.Chem.SDWriter(fobj)
        conf_ids = [conf.GetId() for conf in mol.GetConformers()]
        for i in conf_ids:
            if conf_num not in {-1, None} and i >= conf_num:
                break
            writer.write(mol, confId=i)
        writer.close()
    logging.debug("Saved {:d} conformers to {}.".format(i + 1, out_file))


def mol_to_standardised_mol(mol, name=None):
    """Standardise mol(s)."""
    if name is None:
        try:
            name = mol.GetProp("_Name")
        except:
            name = repr(mol)
    if isinstance(mol, PropertyMol):
        mol_type = PropertyMol
        mol = rdkit.Chem.Mol(mol)
    else:
        mol_type = rdkit.Chem.Mol
    try:
        from standardiser import standardise
        from standardiser.utils import StandardiseException
    except ImportError:
        logging.warning(
            "standardiser module unavailable. Using unstandardised mol.")
        return mol
    logging.debug("Standardising {}".format(name))
    try:
        std_mol = standardise.apply(mol)
        try:
            std_mol.SetProp("_Name", mol.GetProp("_Name"))
        except:
            pass
        return std_mol
    except StandardiseException:
        logging.error(("Standardisation of {} failed. Using unstandardised "
                       "mol.".format(name)), exc_info=True)
    return mol_type(mol)



"""Get E3FP default parameters and read parameters from files.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import os
import copy
import ast
try:
    from ConfigParser import SafeConfigParser, NoSectionError, \
                             DuplicateSectionError
except ImportError:  # Python 3
    from configparser import SafeConfigParser, NoSectionError, \
                             DuplicateSectionError

CONFIG_DIR = os.path.dirname(os.path.realpath(__file__))
DEF_PARAM_FILE = os.path.join(CONFIG_DIR, "defaults.cfg")


def read_params(params=None, fill_defaults=False):
    """Get combination of provided parameters and default parameters.

    Parameters
    ----------
    params : str or SafeConfigParser, optional
        User provided parameters as an INI file or ``SafeConfigParser``.
        Any parameters provided will replace default parameters.
    fill_defaults : bool, optional
        Fill values that aren't provided with package defaults, if `params`
        is file.

    Returns
    -------
    all_params : SafeConfigParser
        Combination of default and user-provided parameters.
    """
    if isinstance(params, SafeConfigParser):
        return copy.copy(params)

    params_list = []
    if fill_defaults:
        params_list.append(DEF_PARAM_FILE)
    if params is not None:
        params_list.append(params)

    all_params = SafeConfigParser()
    all_params.read(params_list)

    return all_params


def write_params(params, params_file="params.cfg"):
    """Write params to file.

    Parameters
    ----------
    params : SafeConfigParser
        Params
    params_file : str
        Params file
    """
    with open(params_file, "w") as f:
        params.write(f)


def get_value(params, section_name, param_name, dtype=str, auto=False,
              fallback=None):
    """Get value from params with fallback.

    Parameters
    ----------
    params : SafeConfigParser
        Parameters
    section_name : str
        Name of section in `params`
    param_name : str
        Name of parameter in `section`
    dtype : type, optional
        Type to return data as.
    auto : bool, optional
        Auto-discover type of value. If provided, `dtype` is ignored.
    fallback : any, optional
        Value to return if getting value fails.

    Returns
    -------
    value : any
        Value of parameter or `fallback`.
    """
    if auto:
        try:
            value = params.get(section_name, param_name)
        except ValueError:
            return fallback

        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            return value
    else:
        get_function = params.get
        if dtype is int:
            get_function = params.getint
        elif dtype is float:
            get_function = params.getfloat
        elif dtype is bool:
            get_function = params.getboolean

        try:
            return get_function(section_name, param_name)
        except ValueError:
            return fallback


def get_default_value(*args, **kwargs):
    global default_params
    return get_value(default_params, *args, **kwargs)


def update_params(params_dict, params=None, section_name=None,
                  fill_defaults=False):
    """Set ``SafeConfigParser`` values from a sections dict.

    Sections dict key must be parameter sections, and value must be dict
    matching parameter name to value. If existing ``SafeConfigParser`` is
    provided, parameter values are updated.

    Parameters
    ----------
    params_dict : dict
        If `section_name` is provided, dict must match parameter names to
        values. If `section_name` is not provided, dict key(s) must be
        parameter sections, and value(s) must be parameter dict.
    params : SafeConfigParser, optional
        Existing parameters.
    section_name : str, optional
        Name of section to which to add parameters in `params_dict`
    fill_defaults : bool, optional
        Fill values that aren't provided with package defaults, if `params`
        is file.
    """
    if params is None:
        params = SafeConfigParser()
    else:
        params = read_params(params, fill_defaults=fill_defaults)

    if section_name is not None:
        try:
            params.add_section(section_name)
        except DuplicateSectionError:
            pass

        for param_name, param_value in params_dict.items():
            params.set(section_name, param_name, str(param_value))
    else:
        sections_dict = params_dict
        for section_name, params_dict in sections_dict.items():
            for param_name, param_value in params_dict.items():
                params.set(section_name, param_name, param_value)
    return params


def params_to_sections_dict(params, auto=True):
    """Get dict of sections dicts in params, with optional type discovery.

    Parameters
    ----------
    params : str or SafeConfigParser
        Params to read
    auto : bool, optional
        Auto typing of parameter values.

    Returns
    ----------
    dict : dict matching sections to parameters to values.
    """
    params = read_params(params)
    sections = default_params.sections()
    params_dicts = {}
    for section in sections:
        try:
            params_dict = dict(params.items(section))
        except NoSectionError:
            continue
        if auto:
            params_dict = {
                param_name: get_value(params, section, param_name, auto=True)
                for param_name in params_dict}
        params_dicts[section] = params_dict
    return params_dicts


default_params = read_params(fill_defaults=True)



"""Functions for generating protonation states of molecules.

Author: Seth Axen
E-mail: seth.axen@gmail.com"""
import os
import tempfile
import subprocess
import itertools
import logging

from .util import iter_to_smiles, MolItemName


def smiles_dict_to_proto_smiles_dict(in_smiles_dict, max_states=3, pka=7.4,
                                     dist_cutoff=20., add_missing=False,
                                     parallelizer=None, chunk_size=100):
    """Generate dict of SMILES for protonated states from SMILES dict."""
    kwargs = {"max_states": max_states, "pka": pka, "dist_cutoff": dist_cutoff}
    in_smiles_iter = ((smiles, mol_name) for mol_name, smiles
                      in in_smiles_dict.items())
    if parallelizer is None:
        proto_smiles_iter = iter(smiles_list_to_proto_smiles_list(
            in_smiles_iter, **kwargs))
    else:
        smiles_chunks_iter = (
            (chunk, ) for chunk in _chunk_iter_to_lists(in_smiles_iter,
                                                        chunk_size=chunk_size))
        results_iter = (
            result for result, data in parallelizer.run_gen(
                smiles_list_to_proto_smiles_list, smiles_chunks_iter,
                kwargs=kwargs) if result is not False)
        proto_smiles_iter = itertools.chain.from_iterable(results_iter)

    proto_smiles_dict = {mol_name: smiles for smiles, mol_name
                         in proto_smiles_iter}
    if add_missing:
        for mol_name, smiles in in_smiles_dict.items():
            proto_name = MolItemName(mol_name, proto_state_num=0).proto_name
            if proto_name not in proto_smiles_dict:
                logging.debug(
                    "Protonated SMILES for %s could not be generated. Returning input SMILES." % mol_name)
                proto_smiles_dict[mol_name] = smiles

    return proto_smiles_dict


def smiles_list_to_proto_smiles_list(in_smiles_list, max_states=3, pka=7.4,
                                     dist_cutoff=20.):
    """Generate list of SMILES for protonated states from single SMILES."""
    in_smiles_file = tempfile.mkstemp(suffix=".smi")[1]
    iter_to_smiles(in_smiles_file,
                   ((mol_name, smiles) for smiles, mol_name in in_smiles_list))
    logging.debug("Protonating SMILES in %s" % (in_smiles_file))
    proc = subprocess.Popen(
        'cxcalc %s --ignore-error dominanttautomerdistribution -H %g -C false -t dist -f "smiles:n,T:dist"' % (
            in_smiles_file, pka), shell=True,
        stdout=subprocess.PIPE)

    proto_smiles_list = []
    try:
        stdout_iter = iter(proc.stdout.readline, b'')
        next(stdout_iter)
        curr_mol_name = None
        curr_states_count = 0
        for line in stdout_iter:
            try:
                smiles, mol_name, dist = line.rstrip('\r\n').split()
            except ValueError:
                logging.warning("Error parsing line:\n%s" % line)
                continue
            if mol_name != curr_mol_name:
                curr_states_count = 0
                curr_mol_name = mol_name
            if curr_states_count >= max_states:
                continue
            if float(dist) > dist_cutoff:
                proto_name = MolItemName(
                    mol_name, proto_state_num=curr_states_count).proto_name
                curr_states_count += 1
                proto_smiles_list.append((smiles, proto_name))
        logging.debug("Finished protonating SMILES in %s" % (in_smiles_file))
    except:
        logging.exception("Error running cxcalc", exc_info=True)

    proc.kill()
    os.remove(in_smiles_file)
    return proto_smiles_list


def smiles_to_proto_smiles(smiles, mol_name, max_states=3, pka=7.4,
                           dist_cutoff=20.):
    """Generate list of SMILES for protonated states from single SMILES.

    This is very inefficient in batch.
    """
    logging.debug("Protonating SMILES in %s" % (mol_name))
    proc = subprocess.Popen(
        'cxcalc "%s %s" --ignore-error dominanttautomerdistribution -H %g -C false -t dist -f "smiles:n,T:dist"' % (
            smiles, mol_name, pka), shell=True,
        stdout=subprocess.PIPE)
    states_count = 0
    proto_smiles_list = []
    try:
        stdout_iter = iter(proc.stdout.readline, b'')
        next(stdout_iter)
        for line in stdout_iter:
            try:
                this_smiles, this_name, dist = line.rstrip('\r\n').split()
            except ValueError:
                logging.warning("Error parsing line:\n%s" % line)
                continue
            if states_count >= max_states:
                break
            if float(dist) > dist_cutoff:
                proto_name = MolItemName(
                    mol_name, proto_state_num=states_count).proto_name
                states_count += 1
                proto_smiles_list.append((smiles, proto_name))
        logging.debug("Finished protonating SMILES in %s" % (mol_name))
    except OSError:
        logging.exception("Error running cxcalc on %s" % (mol_name),
                          exc_info=True)

    proc.kill()
    return proto_smiles_list


def _chunk_iter_to_lists(iterable, chunk_size=100):
    """Yield chunks of size `chunk_size` from iterator."""
    i = 0
    chunk = []
    for item in iterable:
        if i >= chunk_size:
            yield chunk
            chunk = []
            i = 0
        chunk.append(item)
        i += 1
    if len(chunk) != 0:
        yield chunk



"""Tests for E3FP fingerprints.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import unittest


class FingerprintIOTestCases(unittest.TestCase):
    def test_fprint_from_indices(self):
        from e3fp.fingerprint.fprint import Fingerprint, CountFingerprint, \
                                            FloatFingerprint
        for fp_type in (Fingerprint, CountFingerprint, FloatFingerprint):
            in_indices = [3, 1, 4, 5]
            bits = 32
            fprint = fp_type.from_indices(in_indices, bits=bits)
            self.assertEqual(sorted(in_indices), sorted(fprint.indices))

    def test_fprint_from_fprint(self):
        from e3fp.fingerprint.fprint import Fingerprint, CountFingerprint, \
                                            FloatFingerprint
        for fp_type in (Fingerprint, CountFingerprint, FloatFingerprint):
            in_indices = [3, 1, 4, 5, 1, 5, 9]
            bits = 32
            fprint1 = fp_type.from_indices(in_indices, bits=bits)
            fprint2 = fp_type.from_fingerprint(fprint1)
            self.assertEqual(fprint1, fprint2)

    def test_countfprint_from_counts(self):
        from e3fp.fingerprint.fprint import CountFingerprint
        in_counts = {3: 1, 1: 4, 5: 1}
        bits = 32
        fprint = CountFingerprint.from_counts(in_counts, bits=bits)
        out_counts = fprint.counts
        self.assertEqual(in_counts, out_counts)

    def test_floatfprint_from_counts(self):
        from e3fp.fingerprint.fprint import FloatFingerprint
        in_counts = {3: 1.0, 1: 4.0, 5: 1.0}
        bits = 32
        fprint = FloatFingerprint.from_counts(in_counts, bits=bits)
        out_counts = fprint.counts
        self.assertEqual(in_counts, out_counts)

    def test_unique_indices(self):
        from e3fp.fingerprint.fprint import Fingerprint, CountFingerprint, \
                                            FloatFingerprint
        for fp_type in (Fingerprint, CountFingerprint, FloatFingerprint):
            in_indices = [3, 1, 4, 5, 1, 5, 9]
            bits = 32
            fprint = fp_type.from_indices(in_indices, bits=bits)
            self.assertEqual(sorted(set(in_indices)), sorted(fprint.indices))

    def test_bitstring_io(self):
        from e3fp.fingerprint.fprint import Fingerprint, CountFingerprint, \
                                            FloatFingerprint
        for fp_type in (Fingerprint, CountFingerprint, FloatFingerprint):
            in_bitstring = '1001001111011000'
            fprint = fp_type.from_bitstring(in_bitstring)
            out_bitstring = fprint.to_bitstring()
            self.assertEqual(in_bitstring, out_bitstring)

    def test_vector_io(self):
        from e3fp.fingerprint.fprint import Fingerprint, CountFingerprint, \
                                            FloatFingerprint
        import numpy as np
        for fp_type in (Fingerprint, CountFingerprint, FloatFingerprint):
            in_vector = np.array([0, 0, 1, 0, 1, 0, 1, 0, 0],
                                 dtype=np.bool_)
            fprint = fp_type.from_vector(in_vector)
            out_vector = fprint.to_vector(sparse=False)
            np.testing.assert_array_equal(in_vector, out_vector)

    def test_rdkit_io(self):
        from e3fp.fingerprint.fprint import Fingerprint, CountFingerprint, \
                                            FloatFingerprint
        for fp_type in (Fingerprint, CountFingerprint, FloatFingerprint):
            indices = [3, 1, 4, 5]
            bits = 32
            fprint1 = fp_type.from_indices(indices, bits=bits)
            rdkit_fprint1 = fprint1.to_rdkit()
            fprint2 = fp_type.from_rdkit(rdkit_fprint1)
            rdkit_fprint2 = fprint2.to_rdkit()
            self.assertEqual(rdkit_fprint1, rdkit_fprint2)


class FingerprintAlgebraTestCases(unittest.TestCase):
    pass


class FingerprintComparisonTestCases(unittest.TestCase):
    pass

if __name__ == "__main__":
    unittest.main()



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

requirements = [
    'scipy>=0.18.0',
    'numpy>=1.11.3',
    'mmh3>=2.3.1',
    'sdaxen_python_utilities>=0.1.4',
]

test_requirements = ['nose', 'mock']

classifiers = ['Programming Language :: Python',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3.6',
               'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
               'Operating System :: OS Independent',
               'Development Status :: 4 - Beta',
               'Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'Topic :: Scientific/Engineering :: Chemistry',
               'Topic :: Software Development :: Libraries :: Python Modules'
               ]

setup(
    name='e3fp',
    packages=['e3fp', 'e3fp.config', 'e3fp.conformer', 'e3fp.fingerprint',
              'e3fp.test'],
    version='1.0',
    description='Molecular 3D fingerprinting',
    keywords='e3fp 3d molecule fingerprint conformer',
    author='Seth Axen',
    author_email='seth.axen@gmail.com',
    license='LGPLv3',
    url='https://github.com/keiserlab/e3fp',
    classifiers=classifiers,
    download_url='https://github.com/keiserlab/e3fp/tarball/1.0',
    install_requires=requirements,
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=test_requirements,
)



"""Tools for generating E3FP fingerprints.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
from __future__ import division, print_function
import os
import logging

import numpy as np
from rdkit import Chem
import mmh3
from python_utilities.io_tools import touch_dir
from python_utilities.scripting import setup_logging
from ..config.params import get_default_value
from .fprint import Fingerprint, CountFingerprint
from .structs import Shell, shell_to_pdb
from . import array_ops


BITS = 2**32
LEVEL_DEF = get_default_value("fingerprinting", "level", int)
RADIUS_MULTIPLIER_DEF = get_default_value("fingerprinting",
                                          "radius_multiplier", float)
COUNTS_DEF = get_default_value("fingerprinting", "counts", bool)
STEREO_DEF = get_default_value("fingerprinting", "stereo", bool)
INCLUDE_DISCONNECTED_DEF = get_default_value("fingerprinting",
                                             "include_disconnected", bool)
EXCLUDE_FLOATING_DEF = get_default_value("fingerprinting",
                                         "exclude_floating", bool)
IDENT_DTYPE = np.int64  # np.dtype to use for identifiers
Y_AXIS_PRECISION = 0.1  # angstroms
Z_AXIS_PRECISION = 0.01  # rad
POLAR_CONE_RAD = np.pi / 36  # rad
MMH3_SEED = 0
BOND_TYPES = {None: 5, Chem.BondType.SINGLE: 1, Chem.BondType.DOUBLE: 2,
              Chem.BondType.TRIPLE: 3, Chem.BondType.AROMATIC: 4}

setup_logging(reset=False)


class Fingerprinter(object):

    """E3FP fingerprint generator.

    Parameters
    ----------
    bits : int or None, optional
        Maximum number of bits to which to fold returned fingerprint. Multiple
        of 2 is strongly recommended.
    level : int or None, optional
        Maximum number of iterations for fingerprint generation. If None or
        -1, run until no new substructures are identified. Because this could
        produce a different final level number for each conformer, it is
        recommended to manually specify a level.
    radius_multiplier : float, optional
        Multiple by which to increase shell size. At iteration 0, shell radius
        is 0*`radius_multiplier`, at iteration 2, radius is
        2*`radius_multiplier`, etc.
    counts : bool, optional
        Instead of simple bit-based ``Fingerprint`` object, generate
        ``CountFingerprint`` that tracks number of times each bit appears in a
        fingerprint.
    stereo : bool, optional
        Differentiate based on stereography. Resulting fingerprints are not
        comparable to non-stereo fingerprints.
    remove_duplicate_substructs : bool, optional
        If a substructure arises that corresponds to an identifier already in
        the fingerprint, then the identifier for the duplicate substructure is
        not added to fingerprint.
    include_disconnected : bool, optional:
        Include disconnected atoms from hashes and substructure. E3FP's
        advantage over ECFP relies on disconnected atoms, so the option to
        turn this off is present only for testing/comparison.
    exclude_floating : bool, optional:
        Mask atoms with no bonds (usually floating ions) from the fingerprint.
        These are often placed arbitrarily and can confound the fingerprint.

    Attributes
    ----------
    current_level : int
        The maximum level/iteration to which the fingerprinter has been run
        on the current conformer.
    level_shells : dict
        Dict matching level to set of all shells accepted at that level.
    """

    def __init__(self, bits=BITS, level=LEVEL_DEF,
                 radius_multiplier=RADIUS_MULTIPLIER_DEF, stereo=STEREO_DEF,
                 counts=COUNTS_DEF,
                 include_disconnected=INCLUDE_DISCONNECTED_DEF,
                 exclude_floating=EXCLUDE_FLOATING_DEF,
                 remove_duplicate_substructs=True):
        """Initialize fingerprinter settings."""
        self.mol = None
        if level is None:
            level = -1
        self.level = level
        if not np.log2(bits).is_integer():
            logging.warning("bits are not multiple of 2. Multiples of 2 are "
                            "highly recommended")
        self.bits = bits
        self.radius_multiplier = radius_multiplier
        if counts:
            self.fp_type = CountFingerprint
        else:
            self.fp_type = Fingerprint
        self.stereo = stereo
        self.remove_duplicate_substructs = remove_duplicate_substructs

        if self.level == -1 and not self.remove_duplicate_substructs:
            raise Exception(
                "No termination condition specified. 'level' must be "
                "provided or 'remove_duplicate_substructs' must be True")

        self.include_disconnected = include_disconnected
        self.exclude_floating = exclude_floating

        self.bond_types = BOND_TYPES
        self.reset()

    def run(self, conf=None, mol=None, return_substruct=False):
        """Generate fingerprint from provided conformer or mol and conf id.

        Parameters
        ----------
        conf : RDKit Conformer or int, optional
            Input conformer or conformer in `mol`.
        mol : RDKit Mol, optional
            Input molecule object, with at least one conformer. If `conf` not
            specified, first conformer is used.
        return_substruct : bool, optional
            Return dict mapping substructure to fingerprint indices. Keys are
            indices, values are list of substructures, represented as a tuple
            of atom indices where the first index is the central atom and the
            remaining indices (within the sphere) are sorted.
        """
        if mol is None:  # mol not provided; get from conf
            try:
                mol = conf.GetOwningMol()
            except AttributeError:  # conf is int ID; use existing mol
                mol = self.mol
        else:
            if not isinstance(conf, Chem.Conformer):
                try:
                    conf = mol.GetConformer(conf)
                except TypeError:  # conf isn't ID either. Fall back to first
                    conf = mol.GetConformer(0)

        if mol is not self.mol:
            self.reset_mol()
            self.initialize_mol(mol)
        elif conf is not self.conf:
            self.reset_conf()

        self.initialize_conformer(conf)

        for i in iter(self):
            pass

    def reset(self):
        """Clear all variables associated with the last run."""
        self.reset_mol()

    def reset_mol(self):
        """Clear all variables associated with the molecule."""
        self.atoms = None
        self.bound_atoms_dict = {}
        self.connectivity = {}
        self.init_identifiers = {}
        self.reset_conf()

    def reset_conf(self):
        """Clear only conformer-specific variables."""
        self.all_shells = []
        self.atom_coords = None
        # self.current_level = None
        self.identifiers_to_shells = {}
        self.level_shells = {}
        self.past_substructs = set()
        self.shells_gen = None

    def initialize_mol(self, mol):
        """Set general properties of `mol` that apply to all its conformers.

        Parameters
        ----------
        mol : RDKit Mol
            Input molecule ``Mol`` object.
        """
        self.mol = mol
        self.atoms = np.array([x.GetIdx() for x in mol.GetAtoms()
                               if x.GetAtomicNum() > 1])  # ignore hydrogens

        if self.exclude_floating and len(self.atoms) > 1:  # ignore floating atoms
            self.atoms = np.array([x.GetIdx() for x in mol.GetAtoms()
                                   if (x.GetAtomicNum() > 1 and
                                       x.GetDegree() > 0)])

        self.bound_atoms_dict = bound_atoms_from_mol(self.mol, self.atoms)
        self.connectivity = {}
        for i, atom1 in enumerate(self.atoms):
            for j, atom2 in enumerate(self.atoms):
                if i <= j:
                    break
                pair = (int(atom1), int(atom2))
                bond = self.mol.GetBondBetweenAtoms(*pair)
                if bond is not None:
                    bond = bond.GetBondType()
                self.connectivity[pair] = self.bond_types[bond]
                self.connectivity[pair[::-1]] = self.connectivity[pair]

        self.initialize_identifiers()

    def initialize_conformer(self, conf):
        """Retrieve atom coordinates and instantiate shells generator.

        Parameters
        ----------
        conf : RDKit Conformer
            Conformer to fingerprint
        """
        self.conf = conf
        self.atom_coords = coords_from_atoms(self.atoms, self.conf)
        self.shells_gen = ShellsGenerator(
            self.conf, self.atoms, radius_multiplier=self.radius_multiplier,
            atom_coords=self.atom_coords,
            include_disconnected=self.include_disconnected,
            bound_atoms_dict=self.bound_atoms_dict)

    def initialize_identifiers(self):
        """Set initial identifiers for atoms."""
        self.init_identifiers = identifiers_from_invariants(self.mol,
                                                            self.atoms)

    def __next__(self):
        """Run next iteration of fingerprinting."""
        if self.current_level is None:
            shells_dict = next(self.shells_gen)
            if self.current_level != 0:
                raise Exception("ShellsGenerator is not at level 0 at start."
                                " This should never happen.")

            for atom, shell in shells_dict.items():
                shell.identifier = self.init_identifiers[atom]
                self.identifiers_to_shells.setdefault(shell.identifier,
                                                      set()).add(shell)
                self.past_substructs.add(shell.substruct)

            level_shells = set(shells_dict.values())
        else:
            # stop if maximum level has been reached
            if self.current_level >= self.level and self.level != -1:
                logging.debug("Hit maximum level")
                raise StopIteration

            # stop if all substructs contain all atoms (there will never
            # be another new substruct), and that substruct has been seen
            if (self.remove_duplicate_substructs and
                all((len(x.substruct.atoms) == len(self.atoms))
                    for x in self.shells_gen.get_shells_at_level(
                        self.current_level).values())):
                logging.debug("Ran out of substructs")
                raise StopIteration

            shells_dict = next(self.shells_gen)

            for atom, shell in shells_dict.items():
                identifier = identifier_from_shell(shell, self.atom_coords,
                                                   self.connectivity,
                                                   self.current_level,
                                                   self.stereo)
                shell.identifier = identifier

            accepted_shells = sorted(shells_dict.values(),
                                     key=self._shell_to_tuple)

            # filter shells that correspond to already seen substructs
            if self.remove_duplicate_substructs:
                unique_substruct_shells = []
                for shell in accepted_shells:
                    if shell.substruct in self.past_substructs:
                        logging.debug(("Shell with identifier {} at level {} is not "
                                       "unique. Removing.").format(
                                       shell.identifier, self.current_level))
                        shell.is_duplicate = True
                        for x in self.past_substructs:
                            if x == shell.substruct:
                                shell.duplicate = x.shell
                                continue
                        continue
                    unique_substruct_shells.append(shell)
                    self.past_substructs.add(shell.substruct)

                accepted_shells = unique_substruct_shells

            # store shells
            for shell in accepted_shells:
                self.identifiers_to_shells.setdefault(shell.identifier,
                                                      set()).add(shell)

            level_shells = self.level_shells[self.current_level - 1].union(
                set(accepted_shells))
            if (len(level_shells) == len(
                    self.level_shells[self.current_level - 1])):
                self.shells_gen.back()
                logging.debug("No new shells added. Convergence reached.")
                raise StopIteration

        self.all_shells.extend(shells_dict.values())
        self.level_shells[self.current_level] = level_shells

    next = __next__

    @staticmethod
    def _shell_to_tuple(shell):
        return (shell.identifier, shell.center_atom)

    def get_shells_at_level(self, level=-1, exact=False, atom_mask=set()):
        """Get set of shells at the specified level.

        Parameters
        ----------
        level : int or None, optional
            Level/iteration
        exact : bool, optional
            Exact level
        atom_mask : int or set of int, optional
            Don't return shells whose substructures contain these atoms.

        Returns
        -------
        set of Shell : Shells at level
        """
        if level in (-1, None) or level not in self.level_shells:
            if exact or len(self.level_shells) == 0:
                raise IndexError("Level {!r} fingerprints have not yet been "
                                 "generated".format(level))

            true_level = self.current_level
            if level in (-1, None):
                if self.level not in (-1, None):
                    level = self.level
                else:
                    level = true_level
        else:
            true_level = level

        shells = self.level_shells[true_level]
        if len(atom_mask) > 0:
            try:
                atom_mask = set(atom_mask)
            except TypeError:
                atom_mask = {atom_mask}
            shells = {x for x in shells
                      if x.substruct.atoms.isdisjoint(atom_mask)}

        return shells

    def get_fingerprint_at_level(self, level=-1, bits=None, exact=False,
                                 atom_mask=set()):
        """Summary

        Parameters
        ----------
        level : int or None, optional
            Level/iteration
        bits : int or None, optional
            Return fingerprints folded to this number of bits. If unspecified,
            defaults to bits set when instantiated.
        exact : bool, optional
            Exact level
        atom_mask : int or set of int, optional
            Don't return shells whose substructures contain these atoms.

        Returns
        -------
        Fingerprint : Fingerprint at level
        """
        if bits in (-1, None):
            bits = self.bits

        shells = self.get_shells_at_level(level=level, exact=exact,
                                          atom_mask=atom_mask)

        identifiers = signed_to_unsigned_int(
            np.array([x.identifier for x in shells], dtype=IDENT_DTYPE))

        fprint = self.fp_type.from_indices(identifiers, level=level)

        return fprint.fold(bits)

    def substructs_to_pdb(self, level=None, bits=None, out_dir='substructs',
                          reorient=True, exact=False):
        """Save all accepted substructs from current level to PDB.

        Parameters
        ----------
        level : int or None, optional
            Level of fingerprinting/number of iterations
        bits : int or None, optional
            Folding level of identifiers
        out_dir : str, optional
            Directory to which to save PDB files.
        reorient : bool, optional
            Reorient substructure to match stereo quadrants.
        """
        shells = self.get_shells_at_level(level=level, exact=exact)

        if bits in (-1, None):
            bits = self.bits

        touch_dir(out_dir)

        out_files = []
        for shell in shells:
            identifier = signed_to_unsigned_int(shell.identifier) % bits
            out_file = os.path.join(out_dir, "{}.pdb.gz".format(identifier))
            shell_to_pdb(self.mol, shell, self.atom_coords,
                         self.bound_atoms_dict, out_file, reorient=reorient)
            out_files.append(out_file)
        return out_files

    @property
    def current_level(self):
        try:
            return self.shells_gen.level
        except AttributeError:
            return None

    def __iter__(self):
        return self


class ShellsGenerator(object):

    """Generate nested ``Shell`` objects from molecule upon request."""

    def __init__(self, conf, atoms, radius_multiplier=0.5,
                 include_disconnected=True, atom_coords=None,
                 bound_atoms_dict=None):
        """Initialize the generator.

        After initialization, the generator can be iterated to generate a
        ``dict`` matching atom ids to that atom's shell at that
        iteration/level.

        Parameters
        ----------
        conf : RDKit Conformer
            Conformer for which to generate shells.
        atoms : list of int
            Atom ids.
        radius_multiplier : float, optional
            Multiplier by which to increase shell radii.
        include_disconnected : bool, optional
            Include atoms not bonded to center atoms within shell.
        atom_coords : dict or None, optional
            Dict matching atom ids to 1-D array of coordinates. If None, will
            be generated.
        bound_atoms_dict : dict or None, optional
            Dict matching atom ids to set of bound atom ids. If None, will be
            generated.
        """
        self.atoms = tuple(atoms)
        self.radius_multiplier = radius_multiplier
        self.shells_dict = {}
        self.include_disconnected = include_disconnected
        self.level = None

        if atom_coords is None:
            atom_coords = coords_from_atoms(self.atoms, conf)
        atom_coords = [atom_coords.get(x) for x in self.atoms]
        self.distance_matrix = array_ops.make_distance_matrix(atom_coords)

        overlap_atoms = [(self.atoms[i], self.atoms[j]) for i, j in
                         zip(*np.where(self.distance_matrix <= array_ops.EPS))
                         if i < j]
        if len(overlap_atoms) > 0:
            logging.warning("Overlapping atoms {} in conformer {} of molecule"
                            " {}. Fingerprinting will continue but is less "
                            "reliable.".format(
                                ", ".join(map(repr, overlap_atoms)),
                                conf.GetId(),
                                conf.GetOwningMol().GetProp("_Name")))

        if not include_disconnected and bound_atoms_dict is None:
            bound_atoms_dict = bound_atoms_from_mol(conf.GetOwningMol(),
                                                    atoms)
        self.bound_atoms_dict = bound_atoms_dict

    def get_match_atoms(self, rad):
        """Get atoms within shell at radius `rad`.

        Parameters
        ----------
        rad : float
            Radius of shell.

        Returns
        -------
        dict : Dict matching atom id to set of ids for other atoms within
               shell
        """
        match_atoms_dict = {x: set() for x in self.atoms}
        atom_pair_indices_list = list(zip(*np.where(
            self.distance_matrix <= rad)))
        for i, j in atom_pair_indices_list:
            if i <= j:
                continue
            atom1, atom2 = (self.atoms[i], self.atoms[j])
            match_atoms_dict[atom1].add(atom2)
            match_atoms_dict[atom2].add(atom1)
        if not self.include_disconnected:
            match_atoms_dict = {k: v.intersection(self.bound_atoms_dict[k])
                                for k, v in match_atoms_dict.items()}
        return match_atoms_dict

    def __next__(self):
        """Get next iteration's ``dict`` of atom shells."""
        if self.level is None:
            self.level = 0
            self.shells_dict[self.level] = {x: Shell(x, radius=0.)
                                            for x in self.atoms}
            return self.shells_dict[self.level]

        self.level += 1
        self.shells_dict[self.level] = {}
        rad = self.level * self.radius_multiplier
        match_atoms_dict = self.get_match_atoms(rad)
        for atom in self.atoms:
            match_atoms = match_atoms_dict[atom]
            last_match_shells = [self.shells_dict[self.level - 1].get(x)
                                 for x in match_atoms]
            last_shell = self.shells_dict[self.level - 1][atom]
            shell = Shell(atom, last_match_shells, radius=rad,
                          last_shell=last_shell)
            self.shells_dict[self.level][atom] = shell
        return self.shells_dict[self.level]

    next = __next__

    def back(self):
        """Back up one iteration."""
        if self.level in {None, 0}:
            return
        del self.shells_dict[self.level]
        self.level -= 1

    def get_shells_at_level(self, level):
        """Get ``dict`` of atom shells at specified level/iteration.

        If not run to `level`, raises ``IndexError``.

        Parameters
        ----------
        level : int
            Level/iteration from which to retrieve shells ``dict``.

        Returns
        -------
        dict: Dict matching atom ids to that atom's ``Shell`` at that level.
        """
        if level not in self.shells_dict:
            raise IndexError(
                "Level {!r} shells have not been generated".format(level))
        return self.shells_dict[level]

    def __iter__(self):
        return self


# Getting atom properties
def coords_from_atoms(atoms, conf):
    """Build ``dict`` matching atom id to coordinates.

    Parameters
    ----------
    atoms : list of int
        Atom ids
    conf : RDKit Conformer
        Conformer from which to fetch coordinates

    Returns
    -------
    dict : Dict matching atom id to 1-D array of coordinates.
    """
    coordinates = [np.array(conf.GetAtomPosition(int(x)), dtype=np.float64)
                   for x in atoms]
    return dict(zip(atoms, coordinates))


def bound_atoms_from_mol(mol, atoms):
    """Build ``dict`` matching atom id to ids of bounded atoms.

    Bound atoms not in `atoms` are ignored.

    Parameters
    ----------
    mol : RDKit Mol
        Input mol
    atoms : list of int
        List of atom IDs

    Returns
    -------
    dict : Dict matching atom id to set of bound atom ids.
    """
    atoms = set(atoms)
    bound_atoms_dict = {}
    for bond in mol.GetBonds():
        atom1, atom2 = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())
        if (atom1 not in atoms) or (atom2 not in atoms):
            continue
        bound_atoms_dict.setdefault(atom1, set()).add(atom2)
        bound_atoms_dict.setdefault(atom2, set()).add(atom1)
    return bound_atoms_dict


# Generating identifiers
def hash_int64_array(array, seed=MMH3_SEED):
    """Hash an int64 array into a 32-bit integer.

    Parameters
    ----------
    array : ndarray of int64
        Numpy array containing integers
    seed : any, optional
        Seed for MurmurHash3.

    Returns
    -------
    int : 32-bit integer
    """
    if not np.issubdtype(array.dtype, IDENT_DTYPE):
        raise TypeError(
            "Provided array has dtype {} not {}".format(array.dtype,
                                                        IDENT_DTYPE.__name__))
    # ensure all hashed integers are positive
    hashed_int = mmh3.hash(array, seed)
    return hashed_int


def signed_to_unsigned_int(a, bits=BITS):
    """Convert `int` between +/-`bits` to an int between 0 and `bits`.

    Parameters
    ----------
    a : int or ndarray of int
        Integer
    bits : int, optional
        Maximum size of int. E.g. 32-bit is 2^32.

    Returns
    -------
    int : unsigned integer
    """
    return (a + bits) % bits


def identifiers_from_invariants(mol, atoms):
    """Initialize ids according to Daylight invariants.

    Parameters
    ----------
    mol : RDKit Mol
        Input molecule
    atoms : list of int
        IDs for atoms in mol for which to generate identifiers.

    Returns
    -------
    ndarray of int64 : initial identifiers for atoms
    """
    identifiers = [hash_int64_array(
        invariants_from_atom(mol.GetAtomWithIdx(int(x)))) for x in atoms]
    atom_to_identifier_dict = dict(zip(atoms, identifiers))
    return atom_to_identifier_dict


def invariants_from_atom(atom):
    """Get seven invariants from atom.

    Invariants used are the six Daylight invariants, plus an indicator of
    whether the atom is in a ring, as detailed in [1].

    References
    ----------
    1. D Rogers, M Hahn. J. Chem. Inf. Model., 2010, 50 (5), pp 742-754
       https://doi.org/10.1021/ci100050t

    Parameters
    ----------
    atom : RDKit Atom
        Input atom

    Returns
    -------
    1-D array if int64: Array of 7 invariants
    """
    num_hs = atom.GetTotalNumHs()
    return np.array([atom.GetTotalDegree() - num_hs, atom.GetTotalValence() -
                     num_hs, atom.GetAtomicNum(), int(atom.GetMass()),
                     atom.GetFormalCharge(), num_hs, int(atom.IsInRing())],
                    dtype=IDENT_DTYPE)


def identifier_from_shell(shell, atom_coords, connectivity, level, stereo):
    """Determine new identifier for a shell at a specific level.

    Parameters
    ----------
    shell : Shell
        Shell for which to determine identifier
    atom_coords : dict
        Dict matching atom ids to coords.
    connectivity : dict
        Dict matching atom id pair tuples to their bond order (5 for unbound).
    level : int
        Level/iteration
    stereo : bool
        Add stereo indicators
    """
    header = [level, shell.last_shell.identifier]
    atom_tuples = atom_tuples_from_shell(shell, atom_coords, connectivity,
                                         stereo)
    flat_atom_tuples = [y for x in atom_tuples for y in x]
    arr = np.array(header + flat_atom_tuples, dtype=IDENT_DTYPE)
    return hash_int64_array(arr)


def atom_tuples_from_shell(shell, atom_coords, connectivity, stereo):
    """Generate sorted atom tuples for neighboring atoms.

    Parameters
    ----------
    shell : Shell
        Shell for which to build atom tuples
    atom_coords : dict
        Dict matching atom ids to coords.
    connectivity : dict
        Dict matching atom id pair tuples to their bond order (5 for unbound).
    stereo : bool
        Add stereo indicators to tuples
    """
    if len(shell.shells) == 0:
        return []

    # create list matching "bond order" to last iteration's atom identifier
    atom_tuples = [
        (connectivity[(shell.center_atom, x.center_atom)], x.identifier, x)
        for x in shell.shells]

    # add stereo indicator
    if stereo:
        atom_tuples.sort(key=_first_two)
        stereo_indicators = stereo_indicators_from_shell(shell, atom_tuples,
                                                         atom_coords)
        atom_tuples = [x[:-1] + (y,) + (x[-1],) for x, y
                       in zip(atom_tuples, stereo_indicators)]

    # final sort
    atom_tuples = [x[:-1] for x in atom_tuples]
    atom_tuples.sort()
    return atom_tuples


def _first_two(xs):
    return (xs[0], xs[1])


# Methods used for stereo indicators
def pick_y(atom_tuples, cent_coords, y_precision=Y_AXIS_PRECISION):
    """Pick a y-coordinate from atom tuples or mean coordinate.

    Parameters
    ----------
    atom_tuples : list of tuple
        Sorted list of atom tuples
    cent_coords : Nx3 array of float
        Coordinates of atoms with center atom at origin.
    y_precision : str, optional
        For mean to be chosen for y-coordinate, it must be at least this
        distance from the origin. Useful when atoms are symmetrical around
        the center atom where a slight shift in any atom results in a very
        different y.

    Returns
    -------
    1x3 array of float or None : y-coordinate
    int or None : index to y-atom, if y was chosen from the atoms.
    """
    y_inds = get_first_unique_tuple_inds(atom_tuples, 1,
                                         assume_sorted=True)
    # select y-axis vector
    if len(y_inds) > 0:  # unique atom could be found
        y = cent_coords[y_inds, :]
        return y, y_inds[0]
    elif len(atom_tuples) == 2:  # exactly two identical atoms
        y = cent_coords[0, :]
        return y, 0
    else:  # no unique atom could be found
        # assign y-axis as vector to mean coordinate
        y = np.mean(cent_coords, axis=0)
        if np.linalg.norm(y) < y_precision:
            return None, None
        else:
            return y, None


def pick_z(connectivity, identifiers, cent_coords, y, long_angle,
           z_precision=Z_AXIS_PRECISION):
    """Pick a z-coordinate orthogonal to `y`.

    Parameters
    ----------
    connectivity : dict
        Dict matching atom id pair tuples to their bond order (5 for unbound).
    identifiers : iterable of int
        Atom identifiers
    cent_coords : Nx3 array of float
        Coordinates of atoms with center atom at origin.
    y : 1x3 array of float
        y-coordinate
    long_angle : Nx1 array of float
        Absolute angle of atoms from orthogonal to `y`.
    z_precision : str, optional
        Minimum difference in `long_angle` between two potential z-atoms.
        Used as a tie breaker to prevent small shift in one atom resulting
        in very different z.

    Returns
    -------
    1x3 array of float or None : z-coordinate
    """
    angle_from_right = sorted(
        zip(np.asarray(long_angle / z_precision, dtype=np.int),
            connectivity,
            identifiers,
            range(len(identifiers))))

    z_angle_inds = get_first_unique_tuple_inds(angle_from_right, 1,
                                               assume_sorted=True)

    if len(z_angle_inds) > 0:
        z_ind = angle_from_right[z_angle_inds[0]][-1]
        z = array_ops.project_to_plane(cent_coords[z_ind, :], y)
        return z
    else:
        return None


def stereo_indicators_from_shell(shell, atom_tuples, atom_coords_dict,
                                 add_transform_to_shell=True):
    """Get ``list`` of ``int``s indicating location of atoms on unit sphere.

    Parameters
    ----------
    shell : Shell
        Shell for which to get stereo indicators.
    atom_tuples : list of tuple
        List of atom tuples.
    atom_coords_dict : dict
        Dict matching atom ids to coords.
    add_transform_to_shell : bool, optional
        Calculate transformation matrix to align coordinates to unit sphere,
        and add to shell.

    Returns
    -------
    list of int : stereo indicators for atoms in `atom_tuples`.
    """
    cent_coord = atom_coords_dict[shell.center_atom]

    y = None
    z = None
    if len(atom_tuples) > 0:
        (connectivity, identifiers, shells) = list(zip(*atom_tuples))

        stereo_indicators = np.zeros((len(atom_tuples),), dtype=IDENT_DTYPE)
        atoms = [x.center_atom for x in shells]
        mask = np.ones(len(atom_tuples), dtype=np.bool)

        cent_coords = np.array([atom_coords_dict.get(x) for x in atoms],
                               dtype=np.float64) - cent_coord
        # mask atom lying on center atom from consideration
        cent_overlap_indices = np.all(cent_coords == np.zeros(3), axis=1)
        mask[cent_overlap_indices] = False

        # pick y based on first unique atom tuple or mean
        y, y_ind = pick_y(atom_tuples, cent_coords)
        if y_ind is not None:
            mask[y_ind] = False  # use to ensure y not chosen for z

        if y is not None:  # y was picked
            # pick z based on closeness to pi/2 from y-axis
            long_angle = np.pi/2 - array_ops.calculate_angles(cent_coords, y)
            long_angle[np.fabs(long_angle) < array_ops.EPS] = 0.  # perfect right angles
            long_sign = np.asarray(np.sign(long_angle), dtype=IDENT_DTYPE)
            long_sign[long_sign == 0] = 1
            long_angle = np.fabs(long_angle)
            tmp_conn = np.array(connectivity, dtype=IDENT_DTYPE)[mask]
            tmp_ident = np.array(identifiers, dtype=IDENT_DTYPE)[mask]
            z = pick_z(tmp_conn, tmp_ident, cent_coords[mask], y,
                       long_angle[mask])

            if z is not None:  # z was picked
                quad_indicators = quad_indicators_from_coords(cent_coords, y,
                                                              y_ind, z,
                                                              long_sign)
                stereo_indicators = quad_indicators

            #  set indicators for atoms near poles to +/-1
            pole_inds = np.pi / 2 - long_angle < POLAR_CONE_RAD
            stereo_indicators[pole_inds] = long_sign[pole_inds]

        stereo_indicators[cent_overlap_indices] = 0
        stereo_indicators = stereo_indicators.flatten().tolist()
    else:
        stereo_indicators = []

    if add_transform_to_shell:
        shell.transform_matrix = array_ops.make_transform_matrix(cent_coord,
                                                                 y, z)

    return stereo_indicators


def quad_indicators_from_coords(cent_coords, y, y_ind, z, long_sign):
    """Create angle indicators for four quadrants in each hemisphere.

    Parameters
    ----------
    cent_coords : Nx3 array of float
        Array of centered coordinates.
    y : 1-D array of float
        Vector lying along y-axis.
    y_ind : int
        Index of `cent_coords` corresponding to `y`.
    z : 1-D array of float
        Vector lying along z-axis
    long_sign : Nx1 array of int
        Array of signs of vectors in `cent_coords` indicating whether they are
        above (+1) or below (-1) the xz-plane.

    Returns
    -------
    Nx1 array of int : Quadrant indicators. Clockwise from `z` around `y`,
                    indicators are 2, 3, 4, 5 for vectors above the xz-plane
                    and -2, -3, -4, -5 for vectors below the xz-plane.
    """
    atom_lats = array_ops.project_to_plane(cent_coords, y)

    with np.errstate(invalid='ignore'):  # y atom_lat should be (0, 0, 0)
        angle_from_z = array_ops.calculate_angles(atom_lats, z, y).flatten()

    if y_ind is not None:
        angle_from_z[y_ind] = 0.  # otherwise, will be nan

    # offset by pi/4 so z-axis isn't an edge case
    lat_angle = array_ops.rotate_angles(angle_from_z, np.pi / 4)

    # create quadrant indicators
    quad_indicators = 2 + np.asarray(lat_angle * 4/(2 * np.pi),
                                     dtype=IDENT_DTYPE)
    quad_indicators *= long_sign
    return quad_indicators


def get_first_unique_tuple_inds(tuples_list, num_ret, ignore=[],
                                assume_sorted=True):
    """Return indices of first `num_ret` unique tuples in a list.

    Only first 2 values of each tuple are considered.

    Parameters
    ----------
    tuples_list : list of tuple
        List of tuples. Only first two unique values are considered.
    num_ret : int
        Maximum number of first unique tuples to return.
    ignore : list, optional
        Indices for tuples not be considered as unique.
    assume_sorted : bool, optional
        If True, assume list is already sorted by tuples.

    Returns
    -------
    tuple of int : List of at most `num_ret` ints indicating index of
                   unique tuples in list.
    """
    unique = {}
    repeated = set(tuples_list[x][:2] for x in ignore)
    for i, tup in enumerate(tuples_list):
        val = tup[:2]
        if val in repeated:
            if assume_sorted and len(unique) == num_ret:
                break  # okay to stop looking
        elif val in unique:
            del unique[val]
            repeated.add(val)
        else:
            unique[val] = i

    return tuple(sorted(unique.values())[:num_ret])
